import inspect
import logging
import os
import subprocess
import time
from typing import List, Optional, Tuple

import typer
from click import Choice
from libcloud.compute.base import Node

from timestep.infra.cloud_management.cloud_instance_controller import (
    CloudInstanceController,
)
from timestep.infra.cluster_management.k3s_cluster_controller import (
    K3sClusterController,
)
from timestep.utils import ssh_connect

# https://cloud-images.ubuntu.com/locator/
DEFAULT_ALLOWED_IMAGES_IDS = [
    "ami-0e7c4f6b17a66658a",  # AWS
]

DEFAULT_ALLOWED_IMAGE_NAMES = [
    # " (SupportedImages) - Ubuntu 24.04 LTS x86_64 LATEST - 20240603-prod-ajeby5guflgu4", # AWS
    "24.04 (LTS) x64",  # DigitalOcean
    "Ubuntu 24.04 LTS",  # Linode
]

# Configure logging
logging.basicConfig(level=logging.INFO)


def get_help_message():
    is_readme_context = inspect.getmodule(inspect.stack()[1][0]) is None

    return """
Timestep AI CLI - free, local-first, open-source AI
""" + (
        """
## Project Structure

```
src/timestep/
│
├── infra/                  # Infrastructure management
│   ├── cloud_management/   # Cloud instance operations
│   │   └── cloud_instance_controller.py
│   │       - Manages cloud instances using Apache Libcloud
│   │
│   ├── cluster_management/ # Kubernetes cluster management
│   │   └── k3s_cluster_controller.py
│   │       - Manages K3s Kubernetes clusters
│   │
│   └── workload_management/ # Workload orchestration
│       └── sky_workload_controller.py
│           - Manages computational workloads using SkyPilot
│
└── pipelines/              # Data and ML pipeline components
    ├── data_engineering/   # Data preparation stage
    │   └── task.yaml
    │
    ├── machine_learning/   # Model development stage
    │   └── task.yaml
    │
    └── model_deployment/   # Model deployment and monitoring
        └── task.yaml
```

**Development Setup**:

```console
$ python3 -m pip install --upgrade pip
$ python3 -m pip install --user pipx
$ python3 -m pipx ensurepath
$ pipx install poetry==1.8.3
$ cp .env.example .env
$ direnv allow # See https://direnv.net/#getting-started
$ make
$ timestep up --dev
```

**Library Setup**:

```console
$ python3 -m pip install --upgrade pip
$ python3 -m pip install --user pipx
$ python3 -m pipx ensurepath
$ pipx install timestep
$ timestep up
```
"""
        if is_readme_context
        else ""
    )


typer_app = typer.Typer(
    help=get_help_message(),
    no_args_is_help=True,
)


@typer_app.callback()
def main():
    "Timestep AI CLI"


@typer_app.command()
def up(
    allowed_image_ids: Optional[List[str]] = typer.Option(
        default=DEFAULT_ALLOWED_IMAGES_IDS, help="Allowed image IDs to filter by"
    ),
    allowed_image_names: Optional[List[str]] = typer.Option(
        default=DEFAULT_ALLOWED_IMAGE_NAMES, help="Allowed image names to filter by"
    ),
    allowed_location_countries: Optional[List[str]] = typer.Option(
        default=None, help="Allowed location countries to filter by"
    ),
    allowed_location_ids: Optional[List[str]] = typer.Option(
        default=None, help="Allowed location IDs to filter by"
    ),
    allowed_location_names: Optional[List[str]] = typer.Option(
        default=None, help="Allowed location names to filter by"
    ),
    clean: bool = typer.Option(default=False, help="Clean up"),
    dev: bool = typer.Option(default=False, help="Development mode"),
    host: str = typer.Option(default="0.0.0.0", help="Host"),
    min_bandwidth: Optional[int] = typer.Option(
        default=None, help="Minimum bandwidth in GB"
    ),  # TODO: Need to check or convert to proper unit
    min_disk: Optional[int] = typer.Option(
        default=10, help="Minimum disk size in GB"
    ),  # TODO: Need to check or convert to proper unit
    min_ram: Optional[int] = typer.Option(
        # None, help="Minimum RAM in MB"
        default=2000,
        help="Minimum RAM in MB",
    ),  # TODO: Need to check or convert to proper unit
    name: str = typer.Option(default="timestep", help="Name"),
    port: int = typer.Option(default=8000, help="Port"),
    ssh_key: str = typer.Option(
        default="~/.ssh/id_ed25519",
        help="Path to the SSH key",
    ),
):
    """
    Start up the Timestep AI platform.
    """

    if clean:
        typer.echo("\nCleaning up...")

        completed_process: subprocess.CompletedProcess[bytes] = subprocess.run(
            ["./scripts/clean.sh"]
        )

        typer.echo(f"\nCompleted process: {completed_process}")

    # Example credentials (these would be securely stored)
    credentials = {
        # "aws": {
        #     'key': os.getenv('AWS_ACCESS_KEY_ID'),
        #     'secret': os.getenv('AWS_SECRET_ACCESS_KEY')
        # },
        # "azure": {
        #     'key_file': os.getenv('AZURE_KEY_FILE'),
        #     'subscription_id': os.getenv('AZURE_SUBSCRIPTION_ID'),
        # },
        "digital_ocean": {"key": os.getenv("DIGITAL_OCEAN_API_KEY")},
        # "dummy": {
        #     'creds': os.getenv('DUMMY_CREDS')
        # },
        # "gcp": {
        #     'project': os.getenv('GCE_PROJECT'),
        #     'user_id': os.getenv('GCE_USER_ID'),
        # },
        "linode": {"key": os.getenv("LINODE_API_KEY")},
        # ... add more providers here
    }

    # Initialize the controller
    controller = CloudInstanceController(credentials)

    try:
        # Specify instance requirements
        instance_specs = {
            "min_bandwidth": min_bandwidth,
            "min_disk": min_disk,
            "min_ram": min_ram,
        }

        # Find matching sizes
        sizes = controller.find_matching_sizes(instance_specs)

        if not sizes:
            raise typer.BadParameter("No matching instance sizes found")

        def choice(item):
            return f"\n{item}"

        def value_proc(value):
            return value

        size_id = typer.prompt(
            "\nSelect a node size id",
            default=sizes[0].id,
            show_choices=True,
            type=Choice([choice(size) for size in sizes]),
            value_proc=value_proc,
        )

        selected_driver = [size.driver for size in sizes if size.id == size_id][0]
        selected_size = [size for size in sizes if size.id == size_id][0]

        # Find matching images
        images = controller.find_matching_images(
            selected_driver,
            allowed_image_ids,
            allowed_image_names,
        )

        if not images:
            raise typer.BadParameter("No matching images found")

        image_id = typer.prompt(
            "\nSelect a node image id",
            default=images[0].id,
            prompt_suffix=":",
            show_choices=True,
            type=Choice(sorted([choice(image) for image in images])),
            value_proc=value_proc,
        )

        selected_image = [image for image in images if image.id == image_id][0]

        # Find matching locations
        locations = controller.find_matching_locations(
            selected_driver,
            allowed_location_countries,
            allowed_location_ids,
            allowed_location_names,
        )

        if not locations:
            raise typer.BadParameter("No matching locations found")

        location_id = typer.prompt(
            "\nSelect a node location id",
            default="sfo3",
            show_choices=True,
            type=Choice(sorted([choice(location) for location in locations])),
            value_proc=value_proc,
        )

        selected_location = [
            location for location in locations if location.id == location_id
        ][0]

        # Confirm selection
        typer.echo("\nSelected Cloud Instance Details:\n")
        typer.echo(f"Image: {selected_image}")
        typer.echo(f"Location: {selected_location}")
        typer.echo(f"Size: {selected_size}")

        if not typer.confirm("\nConfirm these details?"):
            raise typer.Abort()

        instance = controller.get_instance_by_name(
            driver=selected_driver,
            name=name,
        )

        if instance:
            terminate = typer.confirm(
                f"\nInstance {name} already exists. Do you want to terminate it?"
            )

            if terminate:
                controller.terminate_instance(instance.id)
                typer.echo(f"Terminated instance: {instance}")

        # Create a cost-optimized instance
        node = controller.create_instance(
            driver=selected_driver,
            image=selected_image,
            location=selected_location,
            name=name,
            ssh_key=ssh_key,
            size=selected_size,
        )
        print(f"Created instance: {node}")

        nodes_to_ip_addresses: List[Tuple[Node, List[str]]] = (
            selected_driver.wait_until_running([node])
        )

        for node, ip_addresses in nodes_to_ip_addresses:
            typer.echo(f"\nNode {node.name} is running at {ip_addresses}")

        ip = nodes_to_ip_addresses[0][1][0]

        print(f"ip: {ip}")

        SCRIPT = """#!/usr/bin/env bash
        adduser --disabled-password --gecos "" sky
        usermod -aG sudo sky
        echo "sky ALL=(ALL) NOPASSWD:ALL" > /etc/sudoers.d/sky
        mkdir -p /home/sky/.ssh
        cp /root/.ssh/authorized_keys /home/sky/.ssh/authorized_keys
        chown -R sky:sky /home/sky/.ssh
        chmod 700 /home/sky/.ssh
        chmod 600 /home/sky/.ssh/authorized_keys
        usermod -s /bin/bash sky
        usermod -d /home/sky sky
        passwd -d root
        """

        ssh_connect(ip, script=SCRIPT, username="root", ssh_key=ssh_key)

        ips_file = "ips.txt"

        with open(ips_file, "w") as f:
            f.write(ip)
            f.write("\n")
            f.write("127.0.0.1")

    except Exception as e:
        print(f"Error: {e}")

    typer.echo("\nWaiting for 15 seconds...")
    time.sleep(15)

    username = "sky"

    controller = K3sClusterController(
        cluster_config={
            "ip": ip,
            "ips_file": ips_file,
            "ssh_key": ssh_key,
            "username": username,
        }
    )

    try:
        controller.create_cluster()

    except Exception as e:
        print(f"Error: {e}")

    # typer.echo(f"Starting up the Timestep AI platform at http://{host}:{port}...")

    # timestep_serve(
    #     dev=dev,
    #     host=host,
    #     # llamafile_path=llamafile_path,
    #     port=port,
    # )


if __name__ == "__main__":
    typer_app()



import argparse
import os
from pprint import pprint

import mlflow
import torch
from datasets import load_dataset
from mlflow import MlflowClient
from trl import SFTConfig, SFTTrainer
from unsloth import FastVisionModel, is_bfloat16_supported
from unsloth.trainer import UnslothVisionDataCollator

MLFLOW_TRACKING_URI = os.getenv("MLFLOW_TRACKING_URI")

print(f"MLFLOW_TRACKING_URI: {MLFLOW_TRACKING_URI}")

# PYTORCH_CUDA_ALLOC_CONF=expandable_segments:True
os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "expandable_segments:True"

mlflow.set_tracking_uri(MLFLOW_TRACKING_URI)
# mlflow.transformers.autolog()
mlflow.autolog(
    disable=False,
    disable_for_unsupported_versions=False,
    exclude_flavors=None,
    exclusive=False,
    extra_tags=None,
    # log_dataset=True,
    log_input_examples=True,
    log_model_signatures=True,
    log_models=True,
    # log_trace=True,
    silent=False,
)


def print_auto_logged_info(r):
    tags = {k: v for k, v in r.data.tags.items() if not k.startswith("mlflow.")}
    artifacts = [f.path for f in MlflowClient().list_artifacts(r.info.run_id, "model")]
    print(f"run_id: {r.info.run_id}")
    print(f"artifacts: {artifacts}")
    print(f"params: {r.data.params}")
    print(f"metrics: {r.data.metrics}")
    print(f"tags: {tags}")


dtype = (
    None  # None for auto detection. Float16 for Tesla T4, V100, Bfloat16 for Ampere+
)
load_in_4bit = True  # Use 4bit quantization to reduce memory usage. Can be False.
max_seq_length = 2048  # Supports RoPE Scaling interally, so choose any!
# max_seq_length = 4096 # Choose any! We auto support RoPE Scaling internally!

# [1] Get LAION dataset
# url = "https://huggingface.co/datasets/laion/OIG/resolve/main/unified_chip2.jsonl"
# dataset = load_dataset("json", data_files={"train": url}, split="train")
# dataset = load_dataset("json", data_files={"train": url}, split="train[:1%]")
# dataset = load_dataset("HuggingFaceFW/fineweb", name="CC-MAIN-2024-10", split="train", streaming=True)
# dataset = load_dataset("HuggingFaceFW/fineweb", name="CC-MAIN-2024-46", split="train[:1%]")
# dataset = DataChain.from_hf("HuggingFaceFW/fineweb", name="CC-MAIN-2024-46", split="train")
# dataset = load_dataset("unsloth/Radiology_mini", split = "train[:1%]")
# dataset = load_dataset("unsloth/llava-instruct-mix-vsft-mini", split = "train[:1%]")
# dataset = load_dataset("unsloth/LaTeX_OCR", split = "train")
dataset = load_dataset("unsloth/LaTeX_OCR", split="train[:1%]")
# lmms-lab/LLaVA-OneVision-Data


def convert_to_conversation(sample, instruction, image_key="image", text_key="text"):
    conversation = [
        {
            "role": "user",
            "content": [
                {"type": "text", "text": instruction},
                {"type": "image", "image": sample[image_key]},
            ],
        },
        {"role": "assistant", "content": [{"type": "text", "text": sample[text_key]}]},
    ]

    return {"messages": conversation}


dataset = [
    convert_to_conversation(sample, "Write the LaTeX representation for this image.")
    for sample in dataset
]

# 4bit pre quantized models we support for 4x faster downloading + no OOMs.
# fourbit_models = [
# "unsloth/tinyllama-bnb-4bit",  # "unsloth/tinyllama" for 16bit loading
# ]  # More models at https://huggingface.co/unsloth

# [2] Load Mistral model
# model, tokenizer = FastLanguageModel.from_pretrained(
#     dtype=dtype,
#     load_in_4bit=load_in_4bit,
#     max_seq_length=max_seq_length,
#     # model_name=fourbit_models[0],
#     model_name="TinyLlama/TinyLlama-1.1B-Chat-v1.0",
#     # model_name="ibm-fms/Bamba-9B",
#     # token = "hf_...", # use one if using gated models like meta-llama/Llama-2-7b-hf
# )

model, processor = FastVisionModel.from_pretrained(
    # "unsloth/Llama-3.2-11B-Vision-Instruct",
    # "TinyLlama/TinyLlama-1.1B-Chat-v1.0",
    # "llava-hf/llava-onevision-qwen2-0.5b-ov-hf",
    # "HuggingFaceTB/SmolVLM-Instruct",
    # "Qwen/Qwen2-VL-2B-Instruct",
    "HuggingFaceTB/SmolVLM-Instruct",
    load_in_4bit=True,  # Use 4bit to reduce memory use. False for 16bit LoRA.
    max_seq_length=2048,  # Supports RoPE Scaling internally, so choose any!
    use_gradient_checkpointing="unsloth",  # True or "unsloth" for long context
)

# print("model:")
# print(model)


def print_nparams(model):
    """Calculate the total number of model parameters"""
    nparams = sum(p.numel() for p in model.parameters())
    print(f"Number of parameters: {nparams}")


print_nparams(model)

DEFAULT_SYSTEM_MESSAGE = {
    "content": "You are an AI agent acting as a human assistant.",
    "role": "system",
}

tools_template = """
{
  "tools": [
  {% for tool in tools %}
    {
      "function": {
        "description": "{{ tool.function.description }}",
        "name": "{{ tool.function.name }}",
        "parameters": {{ tool.function.parameters | tojson }}
      },
      "type": "{{ tool.type }}"
    }{% if not loop.last %},{% endif %}\n
  {% endfor %}
  ]
}

If you would like to suggest one or more tool calls, please respond in the following format:
{
  "finish_reason": "tool_calls",
  "tool_calls": [
    {
      "arguments": "{\\"parameter_name\\": \\"parameter_value\\"}",
      "id": "call_id",
      "name": "tool_name"
    }
  ]
}
"""

tool_calls_template = """
{
  "finish_reason": "tool_calls",
  "tool_calls": [
  {% for tool_call in message.tool_calls %}
    {
      "arguments": {{ tool_call.function.arguments | tojson }},
      "id": "{{ tool_call.id }}",
      "name": "{{ tool_call.function.name }}"
    }{% if not loop.last %},{% endif %}\n
  {% endfor %}
  ]
}
"""

tool_response_template = """
{
  "content": {{ message.content | tojson }},
  "name": "{{ message.name }}",
  "tool_call_id": "{{ message.tool_call_id }}"
}
"""

content_template = """{% if message['content'] is string %}{{message['content']}}{% else %}{% for line in message['content'] %}{% if line['type'] == 'text' %}{{line['text']}}{% elif line['type'] in ['image'] %}{{ '<image>' }}{% endif %}{% endfor %}{% endif %}"""

# if pretrained_model_name == "TinyLlama-1.1B-Chat-v1.0":
# start_header_id = "<|"
# end_header_id = "|>"
# start_header_id = "<|im_start|>"  # https://github.com/Mozilla-Ocho/llamafile/blob/a8fd4d28c3d2259c98af7035bcdda1a68af6f62c/llama.cpp/server/server.cpp#L492
start_header_id = ""
# end_header_id = ":" + "' '"
end_header_id = ":"

# else:
#   start_header_id = "<|start_header_id|>"
#   end_header_id = "<|end_header_id|>"

role_header_template = (
    # start_header_id + "{{ message.role }}" + end_header_id + "{{ '\n' }}"
    start_header_id
    + "{{ message.role | capitalize }}"
    + end_header_id
)
assistant_generation_role_header_template = (
    # start_header_id + "assistant" + end_header_id + "{{ '\n' }}"
    start_header_id
    + "Assistant"
    + end_header_id
)

# Influenced by:
# https://cookbook.openai.com/examples/how_to_call_functions_with_chat_models
# https://docs.anthropic.com/en/docs/build-with-claude/tool-use
# https://github.com/abetlen/llama-cpp-python/blob/7c4aead82d349469bbbe7d8c0f4678825873c039/llama_cpp/llama_chat_format.py#L3387
# https://github.com/Mozilla-Ocho/llamafile/blob/66a84d8aea2990895fc4f64786406fea64e79197/llama.cpp/server/server.cpp#L480 (need <|im_start|> b/c Mozilla)
# https://github.com/openai/openai-python/blob/120d225b91a8453e15240a49fb1c6794d8119326/chatml.md
# https://gorilla.cs.berkeley.edu/blogs/8_berkeley_function_calling_leaderboard.html#prompt
# https://huggingface.co/blog/unified-tool-use
chat_template = (
    # Configuration and defaults
    "{%- set config = namespace(has_system_message=false, has_tools=false) -%}"
    "{%- set system_messages = messages | selectattr('role', 'equalto', 'system') | list -%}"
    "{%- set config.has_system_message = system_messages | length > 0 -%}"
    # "{%- set config.has_tools = tools is defined and tools | length > 0 -%}"
    "{%- set config.has_tools = tools is not none and tools | length > 0 -%}"
    # Ensure system message exists
    # "{%- if not config.has_system_message -%}"
    # f'{{%- set messages = [{{ "content": "{DEFAULT_SYSTEM_MESSAGE["content"]}", "role": "{DEFAULT_SYSTEM_MESSAGE["role"]}" }}] + messages -%}}'
    # "{%- endif -%}"
    # Process messages
    "{%- for message in messages -%}"
    # "<|{{ message.role }}|>{{ '\n' }}" # "<|start_header_id|>{{ message.role }}<|end_header_id|>{{ '\n' }}"
    # f"{start_header_id}{{ message.role }}{end_header_id}{{ '\n' }}"
    # start_header_id + "{{ message.role }}" + end_header_id + "{{ '\n' }}"
    # TODO: add bos_token if first message?
    "{% if loop.first %}{{ bos_token }}{% endif %}"
    # f"{role_header_template}" + "{{ ' ' }}"
    "{% if message.content is not string and message.content[0]['type'] in ['image'] %}"
    # "{% if message.content is list and message.content[0]['type'] in ['image', 'image_url'] %}"
    f"{role_header_template}"
    "{%- else -%}"
    f"{role_header_template}" + "{{ ' ' }}"
    "{%- endif -%}"
    # System message handling
    "{%- if message.role == 'system' -%}"
    "{{ message.content }}"
    "{%- if config.has_tools -%}"
    "{{ '\n\n' }}You are aware of the following tools in your environment:"
    f"{tools_template}"
    "{%- endif -%}"
    # "{{ eos_token }}{{ '\n' }}" # <|eot_id|>
    # "<|im_end|>{{ '\n' }}"
    "<end_of_utterance>{{ '\n' }}"
    "{%- endif -%}"
    # User message handling
    "{%- if message.role == 'user' -%}"
    # "{{ message.content }}{{ eos_token }}{{ '\n' }}"
    # "{{ message.content }}<|im_end|>{{ '\n' }}"
    # "{{ message.content }}<end_of_utterance>{{ '\n' }}"
    # f"{content_template}<end_of_utterance>{{ '\n' }}"
    f"{content_template}" + "<end_of_utterance>{{ '\n' }}"
    "{%- endif -%}"
    # Assistant message handling
    "{%- if message.role == 'assistant' -%}"
    "{% generation %}"
    "{%- if message.tool_calls | default(false) -%}"
    f"{tool_calls_template}"
    "{%- else -%}"
    "{{ message.content }}"
    "{%- endif -%}"
    "{% endgeneration %}"
    # "{{ eos_token }}{{ '\n' }}"
    # "<|im_end|>{{ '\n' }}"
    "<end_of_utterance>{{ '\n' }}"
    "{%- endif -%}"
    # Tool message handling
    "{%- if message.role == 'tool' -%}"
    f"{tool_response_template}"
    # "{{ eos_token }}{{ '\n' }}"
    # "<|im_end|>{{ '\n' }}"
    "<end_of_utterance>{{ '\n' }}"
    "{%- endif -%}"
    "{%- endfor -%}"
    # Generation prompt
    "{%- if add_generation_prompt -%}"
    # "<|assistant|>{{ '\n' }}" # <|start_header_id|>assistant<|end_header_id|>
    # f"{assistant_generation_role_header_template}" + "{{ ' ' }}"
    f"{assistant_generation_role_header_template}"
    "{%- endif -%}"
)

# chat_template_with_tool_calls = chat_template

# {
#   "chat_template": "<|im_start|>{% for message in messages %}{{message['role'] | capitalize}}{% if message['content'][0]['type'] == 'image' %}{{':'}}{% else %}{{': '}}{% endif %}{% for line in message['content'] %}{% if line['type'] == 'text' %}{{line['text']}}{% elif line['type'] == 'image' %}{{ '<image>' }}{% endif %}{% endfor %}<end_of_utterance>\n{% endfor %}{% if add_generation_prompt %}{{ 'Assistant:' }}{% endif %}"
# }

# chat_template = "{% for message in messages %}{% if loop.first and messages[0]['role'] != 'system' %}{{ '<|im_start|>system\nYou are a helpful AI assistant<|im_end|>\n' }}{% endif %}<|im_start|>{{message['role'] | capitalize}}{% if message['content'] is string %}{{': '}}{{message['content']}}{% else %}{% if message['content'][0]['type'] == 'image' %}{{':'}}{% else %}{{': '}}{% endif %}{% for line in message['content'] %}{% if line['type'] == 'text' %}{{line['text']}}{% elif line['type'] == 'image' %}{{ '<image>' }}{% endif %}{% endfor %}{% endif %}<end_of_utterance>\n{% endfor %}{% if add_generation_prompt %}{{ 'Assistant: ' }}{% endif %}"
# chat_template = "{% for message in messages %}{% if loop.first and messages[0]['role'] != 'system' %}{{ '<|im_start|>System: You are a helpful AI assistant.<end_of_utterance>\n' }}{% endif %}<|im_start|>{{message['role'] | capitalize}}{% if message['content'] is string %}{{': '}}{{message['content']}}{% else %}{% if message['content'][0]['type'] == 'image' %}{{':'}}{% else %}{{': '}}{% endif %}{% for line in message['content'] %}{% if line['type'] == 'text' %}{{line['text']}}{% elif line['type'] == 'image' %}{{ '<image>' }}{% endif %}{% endfor %}{% endif %}<end_of_utterance>\n{% endfor %}{% if add_generation_prompt %}{{ 'Assistant: ' }}{% endif %}"
# chat_template = "<|im_start|>{% for message in messages %}{% if loop.first and messages[0]['role'] != 'system' %}{{ 'System: You are a helpful AI assistant.<end_of_utterance>\n' }}{% endif %}{{message['role'] | capitalize}}{% if message['content'] is string %}{{': '}}{{message['content']}}{% else %}{% if message['content'][0]['type'] in ['image', 'image_url'] %}{{':'}}{% else %}{{': '}}{% endif %}{% for line in message['content'] %}{% if line['type'] == 'text' %}{{line['text']}}{% elif line['type'] in ['image', 'image_url'] %}{{ '<image>' }}{% endif %}{% endfor %}{% endif %}<end_of_utterance>\n{% endfor %}{% if add_generation_prompt %}{{ 'Assistant: ' }}{% endif %}"

messages = []

# messages += [
#     {
#         "role": "user",
#         "content": [
#             {"type": "image"},
#             {"type": "image"},
#             {"type": "text", "text": "Can you describe the two images?"}
#         ]
#     },
# ]

messages += [
    {
        "role": "user",
        "content": [
            {"type": "text", "text": "What's in this image?"},
            {
                "type": "image_url",
                "image_url": {
                    "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
                },
            },
        ],
    }
]

# messages += [{'content': 'Welcome to the chat! Please enter your name to join.', 'role': 'system'}, {'content': 'Hello! How can I help you today?', 'role': 'assistant'}, {'content': "What's 2+2?", 'role': 'user'}]


print("messages:")
pprint(messages)

# Show example of prompt generated from chat_template before setting it
prompt_before = processor.apply_chat_template(
    add_generation_prompt=True,
    conversation=messages,
    # documents=documents,
    # tools=tools,
)

print("=========================================")
print("\nprompt (before):\n")
print(prompt_before)

# Set the chat_template
processor.chat_template = chat_template
processor.tokenizer.chat_template = chat_template

# Show the same example of prompt generated from chat_template after setting it

prompt = processor.apply_chat_template(
    add_generation_prompt=True,
    conversation=messages,
    # documents=documents,
    # tools=tools,
)

print("\nprompt (after):\n")
print(prompt)

# processor.chat_template = chat_template_with_tool_calls
# processor.tokenizer.chat_template = chat_template_with_tool_calls

# prompt = processor.apply_chat_template(
#     add_generation_prompt=True,
#     conversation=messages,
#     # documents=documents,
#     # tools=tools,
# )

# print("\nprompt (with tool calls):\n")
# print(prompt)
print("=========================================")
print()

assert prompt == prompt_before, f"{prompt} != {prompt_before}"

# raise SystemExit

# print("type(processor):")
# print(type(processor))

# print("processor.chat_template (before):")
# print(processor.chat_template)

# print("processor.tokenizer.chat_template (before):")
# print(processor.tokenizer.chat_template)

assert (
    processor.chat_template == processor.tokenizer.chat_template
), f"{processor.chat_template} != {processor.tokenizer.chat_template}"

# Save the chat_template to file
with open("chat_template.txt", "w") as f:
    f.write(processor.chat_template)

# processor.chat_template = chat_template

# print('tokenizer.chat_template (after):')
# print(processor.chat_template)

print("model.config:")
print(model.config)

# print('model.text_config:')
# print(model.config.text_config)

print("model.generation_config:")
print(model.generation_config)

# assert model.config.bos_token_id == tokenizer.bos_token_id, f"{model.config.bos_token_id} != {tokenizer.bos_token_id}"

# try:
#     assert model.config.eos_token_id == tokenizer.eos_token_id, f"{model.config.eos_token_id} != {tokenizer.eos_token_id}"

# except AssertionError as e:
#     print(e)
#     model.config.eos_token_id = tokenizer.eos_token_id

# assert model.config.pad_token_id == tokenizer.pad_token_id, f"{model.config.pad_token_id} != {tokenizer.pad_token_id}"

# assert model.generation_config.bos_token_id == tokenizer.bos_token_id, f"{model.generation_config.bos_token_id} != {tokenizer.bos_token_id}"

# try:
#     assert model.generation_config.eos_token_id == tokenizer.eos_token_id, f"{model.generation_config.eos_token_id} != {tokenizer.eos_token_id}"

# except AssertionError as e:
#     print(e)
#     model.generation_config.eos_token_id = tokenizer.eos_token_id

# assert model.generation_config.pad_token_id == tokenizer.pad_token_id, f"{model.generation_config.pad_token_id} != {tokenizer.pad_token_id}"

# model.save_pretrained_merged(
#     "models/pretrained_merged_model",
#     # processing_class,
#     # tokenizer,
#     processor,
#     save_method="merged_16bit",
# )

# [3] Do model patching and add fast LoRA weights
# model = FastLanguageModel.get_peft_model(
#     model,
#     bias="none",  # Supports any, but = "none" is optimized
#     loftq_config=None,  # We support LoftQ
#     lora_alpha=16,
#     # lora_alpha=32,
#     lora_dropout=0,  # Supports any, but = 0 is optimized
#     max_seq_length=max_seq_length,
#     r=16,
#     # r = 32, # Choose any number > 0 ! Suggested 8, 16, 32, 64, 128
#     random_state=42,
#     target_modules=[
#         # "embed_tokens",
#         "q_proj",
#         "k_proj",
#         "v_proj",
#         "o_proj",
#         "gate_proj",
#         "up_proj",
#         "down_proj",
#         # "lm_head",
#     ],
#     # [NEW] "unsloth" uses 30% less VRAM, fits 2x larger batch sizes!
#     use_gradient_checkpointing="unsloth",  # True or "unsloth" for very long context
#     use_rslora=False,  # We support rank stabilized LoRA
# )

# model.push_to_hub("mjschock/SmolVLM-Instruct", token=os.getenv("HF_TOKEN"))
# processor.push_to_hub("mjschock/SmolVLM-Instruct", token=os.getenv("HF_TOKEN"))
# model.push_to_hub_merged("mjschock/SmolVLM-Instruct", processor, save_method="merged_16bit", token=os.getenv("HF_TOKEN"))

raise SystemExit

model = FastVisionModel.get_peft_model(
    model,
    finetune_vision_layers=True,  # False if not finetuning vision layers
    finetune_language_layers=True,  # False if not finetuning language layers
    finetune_attention_modules=True,  # False if not finetuning attention layers
    finetune_mlp_modules=True,  # False if not finetuning MLP layers
    r=16,  # The larger, the higher the accuracy, but might overfit
    lora_alpha=16,  # Recommended alpha == r at least
    lora_dropout=0,
    bias="none",
    random_state=42,
    use_rslora=False,  # We support rank stabilized LoRA
    loftq_config=None,  # And LoftQ
    # target_modules = "all-linear", # Optional now! Can specify a list if needed
)

# [4] Parse output directory of checkpoints
parser = argparse.ArgumentParser()
parser.add_argument("--output-dir", type=str, default="outputs")
args = parser.parse_args()

# [5] Initialize and train the model using the SFTTrainer
FastVisionModel.for_training(model)  # Enable for training!

trainer = SFTTrainer(
    # args=TrainingArguments(
    args=SFTConfig(
        bf16=is_bfloat16_supported(),
        fp16=not is_bfloat16_supported(),
        gradient_accumulation_steps=4,
        logging_steps=1,
        learning_rate=2e-4,
        # learning_rate=2e-5,
        lr_scheduler_type="linear",
        # max_steps=60,
        # max_steps=30,
        max_steps=3,
        # num_train_epochs = 1, # Set this instead of max_steps for full training runs
        optim="adamw_8bit",
        # optim = "paged_adamw_8bit",
        output_dir=args.output_dir,
        # per_device_train_batch_size=2,
        per_device_train_batch_size=1,
        # report_to="none", # Use this for WandB etc
        report_to=["mlflow"],
        run_name="Training Run",
        # save_steps=10,
        save_steps=1,
        seed=42,
        # warmup_ratio=0.1,
        # warmup_steps=10,
        # warmup_steps=5,
        warmup_steps=1,
        weight_decay=0.01,
        # You MUST put the below items for vision finetuning:
        remove_unused_columns=False,
        dataset_text_field="",
        dataset_kwargs={"skip_prepare_dataset": True},
        dataset_num_proc=4,
        max_seq_length=2048,
    ),
    data_collator=UnslothVisionDataCollator(model, processor),  # Must use!
    # dataset_num_proc=1, # https://github.com/unslothai/unsloth?tab=readme-ov-file#windows-installation
    # dataset_text_field="text",
    # max_seq_length=max_seq_length,
    model=model,
    tokenizer=processor,
    train_dataset=dataset,
)

gpu_stats = torch.cuda.get_device_properties(0)
start_gpu_memory = round(torch.cuda.max_memory_reserved() / 1024 / 1024 / 1024, 3)
max_memory = round(gpu_stats.total_memory / 1024 / 1024 / 1024, 3)
print(f"\nGPU = {gpu_stats.name}. Max memory = {max_memory} GB.")
print(f"{start_gpu_memory} GB of memory reserved.\n")

# with mlflow.start_run() as run:
# trainer_stats = trainer.train()

# try:
#     trainer_stats = trainer.train()

# except torch._dynamo.exc.BackendCompilerFailed as e:
#     print(e)

torch._dynamo.config.disable = True

print("Dynamic compilation disabled")

trainer_stats = trainer.train()

used_memory = round(torch.cuda.max_memory_reserved() / 1024 / 1024 / 1024, 3)
used_memory_for_lora = round(used_memory - start_gpu_memory, 3)
used_percentage = round(used_memory / max_memory * 100, 3)
lora_percentage = round(used_memory_for_lora / max_memory * 100, 3)
print(f"\n{trainer_stats.metrics['train_runtime']} seconds used for training.")
print(
    f"{round(trainer_stats.metrics['train_runtime']/60, 2)} minutes used for training."
)
print(f"Peak reserved memory = {used_memory} GB.")
print(f"Peak reserved memory for training = {used_memory_for_lora} GB.")
print(f"Peak reserved memory % of max memory = {used_percentage} %.")
print(f"Peak reserved memory for training % of max memory = {lora_percentage} %.\n")

# last_active_run = mlflow.last_active_run()
run_id = mlflow.last_active_run().info.run_id
print(f"run_id: {run_id}")

model.save_pretrained("lora_model")  # Local saving
# tokenizer.save_pretrained("lora_model")
processor.save_pretrained("lora_model")

# remove the models/merged_model directory if it exists
# if os.path.exists("models/merged_model"):
# os.rmdir("models/merged_model")
# os.system("rm -rf models/merged_model")

# https://docs.unsloth.ai/basics/saving-and-using-models/saving-to-gguf - Manual Saving
# trainer.model.save_pretrained_merged(
#     "models/merged_model",
#     trainer.processing_class,
#     # maximum_memory_usage=0.5,
#     maximum_memory_usage=0.25,
#     save_method="merged_16bit",
# )

# run_id = last_active_run.info.run_id
# print_auto_logged_info(mlflow.get_run(run_id=run_id))

# subprocess.run(
#     [
#         "python",
#         "llama.cpp/convert_hf_to_gguf.py",
#         "models/merged_model/",
#         "--outfile",
#         # "models/merged_model/tinyllama-1.1B-chat-bnb-4bit-F16.gguf",
#         "models/merged_model/model-bnb-4bit-F16.gguf",
#         "--outtype",
#         "f16",
#     ]
# )

# subprocess.run(
#     [
#         "llama.cpp/build/bin/llama-quantize",
#         # "models/merged_model/tinyllama-1.1B-chat-bnb-4bit-F16.gguf",
#         "models/merged_model/model-bnb-4bit-F16.gguf",
#         # "models/merged_model/tinyllama-1.1B-chat-bnb-4bit-Q4_K_M.gguf",
#         "models/merged_model/model-bnb-4bit-Q4_K_M.gguf",
#         "Q4_K_M",
#     ]
# )

# subprocess.run(
#     [
#         "ls",
#         "-al",
#         "models/merged_model"
#     ]
# )



import logging
import os
import subprocess

from timestep.utils import ssh_connect


class K3sClusterController:
    """Manages K3s Kubernetes cluster operations."""

    def __init__(self, cluster_config):
        """
        Initialize K3s cluster controller.

        Args:
            cluster_config (dict): Cluster configuration parameters
        """
        self.logger = logging.getLogger(__name__)
        self.cluster_config = cluster_config

    def create_cluster(self):
        """
        Create a new K3s cluster.

        Returns:
            bool: Cluster creation status
        """
        subprocess.run(
            [
                "./scripts/deploy_remote_cluster.sh",
                self.cluster_config["ips_file"],
                self.cluster_config["username"],
                os.path.expanduser(self.cluster_config["ssh_key"]),
            ]
        )

        try:
            import sky.check

            sky.check.check(
                clouds=["kubernetes"],
                quiet=False,
                verbose=True,
            )

        except ModuleNotFoundError as e:
            self.logger.error(f"Failed to import sky.check: {e}")

            subprocess.run(["sky", "check", "k8s"])

        subprocess.run(["sky", "show-gpus", "--cloud", "k8s"])

        SCRIPT = """#!/usr/bin/env bash
        helm install mlflow oci://registry-1.docker.io/bitnamicharts/mlflow --atomic --create-namespace --namespace mlflow
        """

        ssh_connect(
            self.cluster_config["ip"],
            script=SCRIPT,
            username=self.cluster_config["username"],
            ssh_key=os.path.expanduser(self.cluster_config["ssh_key"]),
        )

    def delete_cluster(self):
        """
        Delete the existing K3s cluster.

        Returns:
            bool: Cluster deletion status
        """
        raise NotImplementedError()



import os
import time

import paramiko


def ssh_connect(
    ip_address: str, script: str, username: str, ssh_key: str
) -> paramiko.SSHClient:
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    max_attempts = 5
    for attempt in range(max_attempts):
        try:
            client.connect(
                hostname=ip_address,
                username=username,
                key_filename=os.path.expanduser(ssh_key),
                timeout=10,
            )

            _, stdout, stderr = client.exec_command(script)

            print("stdout:")
            print(stdout.read().decode())

            print("stderr:")
            print(stderr.read().decode())

            return client

        except Exception as e:
            print(f"Attempt {attempt + 1} failed: {e}")
            time.sleep(10)

    raise Exception("Failed to connect after multiple attempts")



from libcloud.compute.base import KeyPair, NodeDriver


def get_or_create_key_pair(node_driver: NodeDriver, name: str, content: str) -> KeyPair:
    # TODO: what about we want to delete the key pair? delete_key_pair

    key_pair: KeyPair | None = node_driver.get_key_pair(name=name)

    if not key_pair:
        key_pair: KeyPair = node_driver.create_key_pair(name=name, public_key=content)

    return key_pair



import asyncio
import json
import os
from contextlib import asynccontextmanager
from enum import Enum

import flet as ft
import flet.fastapi as flet_fastapi
import uvicorn
from connexion import AsyncApp, ConnexionMiddleware
from connexion.options import SwaggerUIOptions
from fastapi import FastAPI, Request
from openai import AsyncOpenAI

from timestep.config import settings

app_dir = settings.app_dir
connexion_app = AsyncApp(import_name=__name__)
default_hf_repo_id = settings.default_hf_repo_id

client = AsyncOpenAI(
    api_key="sk-no-key-required",
    # base_url="http://localhost:8080/v1",
    base_url="http://localhost:8000/v1",
)


@asynccontextmanager
async def lifespan(app: FastAPI):
    try:
        print("=== Lifespan startup ===")

        await flet_fastapi.app_manager.start()
        yield
        await flet_fastapi.app_manager.shutdown()

    finally:
        print("=== Lifespan cleanup ===")
        print("Shutting down...")


fastapi_app = FastAPI(
    lifespan=lifespan,
    # servers= # TODO: add server for testing?
)

# connexion_app.add_api(
#     "api/openai/v1/openapi/openapi.yaml",
#     base_path="/openai/v1",
#     pythonic_params=True,
#     resolver_error=500,
# )

fastapi_app.mount("/api", ConnexionMiddleware(app=connexion_app, import_name=__name__))


class Message:
    def __init__(self, user_name: str, text: str, message_type: str):
        self.user_name = user_name
        self.text = text
        self.message_type = message_type


class ChatMessage(ft.Row):
    def __init__(self, message: Message):
        super().__init__()

        self.user_name_view = ft.Text(message.user_name, weight="bold")
        self.text_view = ft.Text(message.text, selectable=True)

        self.controls = [
            ft.CircleAvatar(
                content=ft.Text(self.get_initials(message.user_name)),
                color=ft.colors.WHITE,
                bgcolor=self.get_avatar_color(message.user_name),
            ),
            ft.Column(
                [
                    # ft.Text(message.user_name, weight="bold"),
                    self.user_name_view,
                    # ft.Text(message.text, selectable=True),
                    # ft.Text(self.text, selectable=True),
                    self.text_view,
                ],
                tight=True,
                spacing=5,
            ),
        ]
        self.vertical_alignment = ft.CrossAxisAlignment.START

    def get_initials(self, user_name: str):
        if user_name:
            return user_name[:1].capitalize()

        else:
            return "Unknown"  # or any default value you prefer

    def get_avatar_color(self, user_name: str):
        colors_lookup = [
            ft.colors.AMBER,
            ft.colors.BLUE,
            ft.colors.BROWN,
            ft.colors.CYAN,
            ft.colors.GREEN,
            ft.colors.INDIGO,
            ft.colors.LIME,
            ft.colors.ORANGE,
            ft.colors.PINK,
            ft.colors.PURPLE,
            ft.colors.RED,
            ft.colors.TEAL,
            ft.colors.YELLOW,
        ]

        return colors_lookup[hash(user_name) % len(colors_lookup)]

    def to_dict(self):
        if self.user_name_view.value == "Assistant":
            role = "assistant"

        elif self.user_name_view.value == "System":
            role = "system"

        else:
            role = "user"

        return {
            "content": self.text_view.value,
            "role": role,
        }

    def update_text(self, text: str):
        self.text_view.value = text
        self.update()


# async def main(page: ft.Page):
async def root_main(page: ft.Page):
    page.horizontal_alignment = ft.CrossAxisAlignment.STRETCH
    page.title = "Chat"

    def join_chat_click(e):
        if not join_user_name.value:
            join_user_name.error_text = "Name cannot be blank!"
            join_user_name.update()

        else:
            page.session.set("user_name", join_user_name.value)
            page.dialog.open = False
            new_message.prefix = ft.Text(f"{join_user_name.value}: ")
            page.pubsub.send_all(
                Message(
                    user_name=join_user_name.value,
                    text=f"{join_user_name.value} has joined the chat.",
                    message_type="login_message",
                )
            )
            page.update()

    def send_message_click(e):
        if new_message.value != "":
            page.pubsub.send_all(
                Message(
                    page.session.get("user_name"),
                    new_message.value,
                    message_type="chat_message",
                )
            )
            new_message.value = ""
            new_message.focus()
            page.update()

    async def on_message(message: Message):
        if message.message_type == "chat_message":
            m = ChatMessage(message)

        elif message.message_type == "login_message":
            # m = ft.Text(message.text, italic=True, color=ft.colors.BLACK45, size=12)
            m = ft.Text(message.text, italic=True, color=ft.colors.WHITE, size=12)

        chat.controls.append(m)
        page.update()

        if message.message_type == "chat_message" and not message.user_name in (
            "Assistant",
            "System",
        ):
            messages = [
                chat_message.to_dict()
                for chat_message in chat.controls
                if type(chat_message) == ChatMessage
            ]

            print("messages:")
            print(messages)

            assistant_message = ChatMessage(
                Message(
                    user_name="Assistant",
                    text="...",
                    message_type="chat_message",
                )
            )

            chat.controls.append(assistant_message)
            page.update()

            stream = await client.chat.completions.create(
                # max_tokens=100,
                max_completion_tokens=100,
                # messages=[{"role": "user", "content": "Say this is a test"}],
                messages=messages,
                # model="LLaMA_CPP",
                model="HuggingFaceTB/SmolVLM-Instruct",
                stop=[
                    # "<|assistant|>",
                    # "<|endoftext|>",
                    # "<|system|>",
                    # "<|user|>",
                    # "</s>",
                    # "<end_of_utterance>",
                ],
                stream=True,
            )

            chunks = []

            async for chunk in stream:
                # print(chunk.choices[0].delta.content or "", end="")
                # assistant_message.update_text(chunk.choices[0].delta.content or "")
                chunks.append(chunk.choices[0].delta.content or "")

                assistant_message.update_text("".join(chunks))

            # assistant_message.update_text(chat_completion.choices[0].message)

    page.pubsub.subscribe(on_message)

    # A dialog asking for a user display name
    join_user_name = ft.TextField(
        label="Enter your name to join the chat",
        autofocus=True,
        on_submit=join_chat_click,
    )

    page.dialog = ft.AlertDialog(
        open=True,
        modal=True,
        title=ft.Text("Welcome!"),
        content=ft.Column([join_user_name], width=300, height=70, tight=True),
        actions=[ft.ElevatedButton(text="Join chat", on_click=join_chat_click)],
        actions_alignment=ft.MainAxisAlignment.END,
    )

    # Chat messages
    chat = ft.ListView(
        expand=True,
        spacing=10,
        auto_scroll=True,
    )

    # Add a welcome message
    message = Message(
        user_name="System",
        text="Welcome to the chat! Please enter your name to join.",
        message_type="chat_message",
    )
    await on_message(message)

    message = Message(
        user_name="Assistant",
        text="Hello! How can I help you today?",
        message_type="chat_message",
    )
    await on_message(message)

    # A new message entry form
    new_message = ft.TextField(
        hint_text="Write a message...",
        autofocus=True,
        shift_enter=True,
        min_lines=1,
        max_lines=5,
        filled=True,
        expand=True,
        on_submit=send_message_click,
    )

    # Add everything to the page
    page.add(
        ft.Container(
            content=chat,
            border=ft.border.all(1, ft.colors.OUTLINE),
            border_radius=5,
            padding=10,
            expand=True,
        ),
        ft.Row(
            [
                new_message,
                ft.IconButton(
                    icon=ft.icons.SEND_ROUNDED,
                    tooltip="Send message",
                    on_click=send_message_click,
                ),
            ]
        ),
    )


fastapi_app.mount("/", flet_fastapi.app(root_main))


class Engine(str, Enum):
    copilot_codex = "copilot-codex"


@fastapi_app.get("/v1/engines")
async def list_engines():
    # return [engine.value for engine in Engine]
    return {
        "data": [
            {
                "id": Engine.copilot_codex,
                "name": "Copilot Codex",
                "description": "OpenAI's Codex model, formerly known as GitHub Copilot",
            }
        ]
    }


@fastapi_app.post("/v1/engines/{engine}/completions")
async def create_code_completion(engine: Engine, request: Request):
    print(
        f"=== {__name__}.create_code_completion(engine: Engine, request: Request) ==="
    )
    print(f"engine: {engine}")
    print(f"request: {request}")

    if engine == Engine.copilot_codex:
        try:
            body = await request.json()

        except json.decoder.JSONDecodeError as e:
            print(f"Error: {e}")
            body = {
                "model": "LLaMA_CPP",
                "prompt": "int main() {",
            }

        # body["model"] = "LLaMA_CPP"

        # return await create_completion(body, token_info={}, user=None)
        raise NotImplementedError("Not implemented")

    else:
        raise NotImplementedError("Not implemented")


def main(*args, **kwargs):
    print(f"=== {__name__}.main(*args, **kwargs) ===")
    print(f"args: {args}")
    print(f"kwargs: {kwargs}")

    cwd = os.getcwd()
    print(f"cwd: {cwd}")

    uvicorn.run(
        f"{__name__}:fastapi_app",
        # host="0.0.0.0",
        host=kwargs.get("host", "0.0.0.0"),
        log_level="info",
        loop="asyncio",
        # port=8000,
        # port=kwargs.get("port", 8000),
        port=kwargs.get("port", 8080),
        # reload=True,
        reload=kwargs.get("dev", False),
        reload_dirs=[
            # f"{os.getcwd()}/timestep",
            f"{os.getcwd()}/src/timestep",
        ],
        workers=1,
    )


if __name__ == "__main__":
    main()



import logging
import os
from typing import Dict, List, Optional

from libcloud.compute.base import (
    KeyPair,
    Node,
    NodeDriver,
    NodeImage,
    NodeLocation,
    NodeSize,
)
from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

from timestep.infra.cloud_management.utils import get_or_create_key_pair


class CloudInstanceController:
    """
    Manages cloud instances by finding the most cost-effective option
    across multiple cloud providers.
    """

    def __init__(self, credentials: Dict[str, Dict]):
        """
        Initialize cloud instance controller with credentials for multiple providers.

        Args:
            credentials (Dict[str, Dict]): A dictionary of cloud provider credentials
                Example: {
                    'aws': {'key': 'aws_access_key', 'secret': 'aws_secret_key'},
                    'digital_ocean': {'key': 'do_access_key'},
                    'gcp': {'key': 'gcp_service_account_key'}
                }
        """
        self.providers = {}
        self.logger = logging.getLogger(__name__)

        # Map of cloud providers supported by libcloud
        provider_map = {
            "aws": Provider.EC2,
            "azure": Provider.AZURE,
            "digital_ocean": Provider.DIGITAL_OCEAN,
            "dummy": Provider.DUMMY,
            "gcp": Provider.GCE,
            "linode": Provider.LINODE,
        }

        # Initialize drivers for each provider
        for provider, creds in credentials.items():
            try:
                driver_class = get_driver(provider_map.get(provider))

                self.providers[provider] = driver_class(**creds)
                self.logger.info(f"Initialized {provider} cloud driver")

            except Exception as e:
                self.logger.error(f"Failed to initialize {provider} driver: {e}")

    def create_instance(
        self,
        driver: NodeDriver,
        image: NodeImage,
        location: NodeLocation,
        name: str,
        ssh_key: str,
        size: NodeSize,
    ) -> Node:
        """
        Create the most cost-effective cloud instance matching specified requirements.

        Args:
            driver (NodeDriver): Cloud provider driver
            image (NodeImage): Image to use for the instance
            location (NodeLocation): Location to deploy the instance
            name (str): Name of the instance
            ssh_key (str): SSH key to use for the instance
            size (NodeSize): Instance size specifications

        Returns:
            Node: Created instance object
        """
        try:
            key_file_path = os.path.expanduser(f"{ssh_key}.pub")

            with open(key_file_path) as fp:
                key_material = fp.read()

            key_pair: KeyPair = get_or_create_key_pair(driver, name, key_material)

            node = driver.create_node(
                ex_create_attr={"ssh_keys": [key_pair.fingerprint]},
                name=name,
                image=image,
                location=location,
                size=size,
            )

            self.logger.info(f"Created instance: {node}")

            return node

        except Exception as e:
            self.logger.error(f"Failed to create instance: {e}")
            raise

    def find_matching_images(
        self,
        driver: NodeDriver,
        allowed_image_ids: List[str] = None,
        allowed_image_names: List[str] = None,
    ) -> List[NodeImage]:
        """
        Find images across providers that match allowed IDs or names.

        Args:
            driver (NodeDriver): Cloud provider driver
            allowed_image_ids (List[str], optional): Specific image IDs to match
            allowed_image_names (List[str], optional): Name patterns to match

        Returns:
            List[NodeImage]: List of matching images
        """
        matching_images = []

        try:
            images = driver.list_images()

            # Filter images based on IDs and names
            provider_matches = []

            for image in images:
                # Check if image matches any allowed ID
                id_match = (not allowed_image_ids) or (image.id in allowed_image_ids)

                # Check if image matches any name pattern
                name_match = (not allowed_image_names) or any(
                    name.lower() == str(image.name).lower()
                    for name in allowed_image_names
                )

                if id_match or name_match:
                    provider_matches.append(image)

            matching_images.extend(provider_matches)

        except Exception as e:
            self.logger.warning(f"Error finding images for {driver}: {e}")

        return matching_images

    def find_matching_locations(
        self,
        driver: NodeDriver,
        allowed_location_countries: List[str] = None,
        allowed_location_ids: List[str] = None,
        allowed_location_names: List[str] = None,
    ) -> List[NodeLocation]:
        """
        Find locations across providers, optionally filtering by preferred regions.

        Args:
            driver (NodeDriver): Cloud provider driver
            allowed_location_countries (List[str], optional): Country names to match
            allowed_location_ids (List[str], optional): Specific location IDs to match
            allowed_location_names (List[str], optional): Name patterns to match

        Returns:
            List[NodeLocation]: List of matching locations
        """
        matching_locations = []

        try:
            locations = driver.list_locations()

            # Filter locations if preferred regions specified
            provider_matches = []

            for location in locations:
                # Check if image matches any allowed ID
                country_match = (not allowed_location_countries) or any(
                    country.lower() == str(location.country).lower()
                    for country in allowed_location_countries
                )

                id_match = (not allowed_location_ids) or (
                    location.id in allowed_location_ids
                )

                # Check if image matches any name pattern
                name_match = (not allowed_location_names) or any(
                    name.lower() == str(location.name).lower()
                    for name in allowed_location_names
                )

                if country_match or id_match or name_match:
                    provider_matches.append(location)

            matching_locations.extend(provider_matches)

        except Exception as e:
            self.logger.warning(f"Error finding locations for {driver}: {e}")

        return matching_locations

    def find_matching_sizes(self, specs: Dict, limit: int = 3) -> List[NodeSize]:
        """
        Retrieve and compare instance pricing across providers.

        Args:
            specs (Dict): Desired instance specifications
                {
                    'min_bandwidth': int,  # Minimum bandwidth in Mbps
                    'min_disk': int,  # Minimum disk size in GB
                    'min_ram': int,  # Minimum RAM in MB
                }

        Returns:
            List[NodeSize]: List of matching instance sizes
        """
        available_instances = []

        for provider_name, driver in self.providers.items():
            try:
                # Fetch available nodes matching specifications
                sizes = driver.list_sizes()

                for size in sizes:
                    if (
                        (
                            specs["min_bandwidth"] is None
                            or size.bandwidth >= specs["min_bandwidth"]
                        )
                        and (
                            specs["min_disk"] is None or size.disk >= specs["min_disk"]
                        )
                        and size.price > 0
                        and (specs["min_ram"] is None or size.ram >= specs["min_ram"])
                    ):
                        available_instances.append(
                            size,
                        )

            except Exception as e:
                self.logger.warning(
                    f"Error fetching instances for {provider_name}: {e}"
                )

        # Sort instances by hourly cost
        return sorted(available_instances, key=lambda x: int(x.price))[0:limit]

    def get_instance_by_name(self, driver: NodeDriver, name: str) -> Optional[Node]:
        """
        Retrieve a cloud instance by name.

        Args:
            name (str): Name of the instance to retrieve

        Returns:
            Node: Instance object if found, else None
        """
        try:
            nodes = driver.list_nodes()
            return next((node for node in nodes if node.name == name), None)

        except Exception as e:
            self.logger.error(f"Error fetching instance by name: {e}")
            return None

    def terminate_instance(self, instance_id: str, provider: Optional[str] = None):
        """
        Terminate a specific cloud instance.

        Args:
            instance_id (str): Unique identifier of the instance to terminate
            provider (Optional[str]): Specific provider (if known)
        """
        if provider:
            # If provider is specified, use its driver
            driver = self.providers.get(provider)
            if not driver:
                raise ValueError(f"Provider {provider} not initialized")

            nodes = driver.list_nodes()
            target_node = next((node for node in nodes if node.id == instance_id), None)

            if target_node:
                driver.destroy_node(target_node)
                self.logger.info(f"Terminated instance {instance_id} on {provider}")
            else:
                raise ValueError(f"Instance {instance_id} not found on {provider}")

        else:
            # If no provider specified, search across all providers
            for provider_name, driver in self.providers.items():
                try:
                    nodes = driver.list_nodes()
                    target_node = next(
                        (node for node in nodes if node.id == instance_id), None
                    )

                    if target_node:
                        driver.destroy_node(target_node)
                        self.logger.info(
                            f"Terminated instance {instance_id} on {provider_name}"
                        )
                        return

                except Exception as e:
                    self.logger.warning(f"Error searching {provider_name}: {e}")

            raise ValueError(f"Instance {instance_id} not found on any provider")



import asyncio
import json
import logging
import os
import time
from pprint import pprint
from threading import Thread
from typing import Any, Dict, List

import torch
from fastapi import FastAPI, Request
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion import Choice as ChatCompletionChoice
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
from openai.types.chat.chat_completion_chunk import Choice as ChatCompletionChunkChoice
from openai.types.chat.chat_completion_chunk import ChoiceDelta
from openai.types.chat.chat_completion_message import ChatCompletionMessage
from openai.types.chat.completion_create_params import CompletionCreateParams
from pydantic import TypeAdapter
from ray import serve
from sse_starlette import EventSourceResponse
from starlette.responses import JSONResponse
from transformers import TextIteratorStreamer

# from transformers.generation.streamers import AsyncTextIteratorStreamer
from transformers.image_utils import load_image
from unsloth import FastVisionModel

dtype = (
    None  # None for auto detection. Float16 for Tesla T4, V100, Bfloat16 for Ampere+
)
load_in_4bit = True  # Use 4bit quantization to reduce memory usage. Can be False.
max_seq_length = 2048  # Supports RoPE Scaling interally, so choose any!
# max_seq_length = 4096 # Choose any! We auto support RoPE Scaling internally!


logger = logging.getLogger("ray.serve")

os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "expandable_segments:True"

app = FastAPI()

# middlewares = [
#     middleware
#     for middleware in ConnexionMiddleware.default_middlewares
#     if middleware is not SecurityMiddleware
# ]

# connexion_app = AsyncApp(import_name=__name__, middlewares=middlewares)

# connexion_app.add_api(
#     # "api/openai/v1/openapi/openapi.yaml",
#     "api/v1/openapi/openapi.yaml",
#     # base_path="/openai/v1",
#     base_path="/v1",
#     pythonic_params=True,
#     resolver_error=501,
# )

# # fastapi_app.mount("/api", ConnexionMiddleware(app=connexion_app, import_name=__name__))
# # app.mount("/api", ConnexionMiddleware(app=connexion_app, import_name=__name__))
# app.mount(
#     "/",
#     ConnexionMiddleware(
#         app=connexion_app,
#         import_name=__name__,
#         # middlewares=middlewares,
#     ),
# )


@serve.deployment(
    autoscaling_config={
        # https://docs.ray.io/en/latest/serve/advanced-guides/advanced-autoscaling.html#required-define-upper-and-lower-autoscaling-limits
        "max_replicas": 1,
        "min_replicas": 1,  # TOOD: set to 0
        "target_ongoing_requests": 2,  # https://docs.ray.io/en/latest/serve/advanced-guides/advanced-autoscaling.html#target-ongoing-requests-default-2
    },
    max_ongoing_requests=5,  # https://docs.ray.io/en/latest/serve/advanced-guides/advanced-autoscaling.html#max-ongoing-requests-default-5
    ray_actor_options={"num_gpus": 1},
)
@serve.ingress(app)
class VLLMDeployment:
    def __init__(
        self,
        model_name: str,
    ):
        self.model_name = model_name

        model, processor = FastVisionModel.from_pretrained(
            load_in_4bit=load_in_4bit,
            max_seq_length=max_seq_length,
            model_name=self.model_name,
        )

        with open("chat_template.txt", "r") as f:
            processor.chat_template = f.read()
            processor.tokenizer.chat_template = processor.chat_template

        FastVisionModel.for_inference(model)  # Enable native 2x faster inference

        self.model = model
        self.processor = processor

    def reconfigure(self, config: Dict[str, Any]):
        print("=== reconfigure ===")
        print("config:")
        print(config)
        # https://docs.ray.io/en/latest/serve/production-guide/config.html#dynamically-change-parameters-without-restarting-replicas-user-config

    @app.post("/v1/chat/completions")
    async def create_chat_completion(self, body: dict, raw_request: Request):
        """Creates a model response for the given chat conversation. Learn more in the [text generation](/docs/guides/text-generation), [vision](/docs/guides/vision), and [audio](/docs/guides/audio) guides.  Parameter support can differ depending on the model used to generate the response, particularly for newer reasoning models. Parameters that are only supported for reasoning models are noted below. For the current state of  unsupported parameters in reasoning models,  [refer to the reasoning guide](/docs/guides/reasoning).

        # noqa: E501

        :param create_chat_completion_request:
        :type create_chat_completion_request: dict | bytes

        :rtype: Union[CreateChatCompletionResponse, Tuple[CreateChatCompletionResponse, int], Tuple[CreateChatCompletionResponse, int, Dict[str, str]]
        """
        print("=== create_chat_completion ===")

        print("body:")
        pprint(body)

        ta = TypeAdapter(CompletionCreateParams)

        print("ta.validate_python...")
        pprint(ta.validate_python(body))

        max_new_tokens = body.get("max_completion_tokens", body.get("max_tokens"))
        messages = body.get("messages")
        model_name = body.get("model")
        stream = body.get("stream", False)
        temperature = body.get("temperature")
        tools = body.get("tools")

        images = []

        for message in messages:
            for content in message["content"]:
                if "type" in content and content["type"] == "image_url":
                    image_url = content["image_url"]["url"]
                    image = load_image(image_url)
                    images.append(image)

                    content["type"] = "image"
                    del content["image_url"]

        images = images if images else None

        if model_name != self.model_name:
            return JSONResponse(content={"error": "Model not found"}, status_code=404)

        prompt = self.processor.apply_chat_template(
            add_generation_prompt=True,
            conversation=messages,
            # documents=documents,
            temperature=temperature,
            tools=tools,
        )

        print("prompt:")
        print(prompt)

        print("images:")
        print(images)

        inputs = self.processor(text=prompt, images=images, return_tensors="pt")
        inputs = inputs.to(self.model.device)
        input_ids = inputs.input_ids

        class GeneratorThread(Thread):
            """Thread to generate completions in the background."""

            def __init__(self, model, **generation_kwargs):
                super().__init__()

                self.chat_completion = None
                self.generation_kwargs = generation_kwargs
                self.model = model

            def run(self):
                import torch
                import torch._dynamo.config

                try:
                    try:
                        self.generated_ids = self.model.generate(
                            **self.generation_kwargs
                        )

                    except torch._dynamo.exc.BackendCompilerFailed as e:
                        print(e)
                        print("Disabling dynamo...")

                        # TODO: Use CompileConfig to disable dynamo?
                        # https://huggingface.co/docs/transformers/v4.47.1/en/internal/generation_utils#transformers.CompileConfig
                        # compile_config = CompileConfig(dynamic=True)
                        # If this works, I can skip this GeneratorThread class and just call thread = Thread(target=model.generate, kwargs=generation_kwargs)
                        torch._dynamo.config.disable = True

                        self.generated_ids = self.model.generate(
                            **self.generation_kwargs
                        )

                except Exception as e:
                    print(e)
                    print("Warning: Exception in GeneratorThread")
                    self.generated_ids = []

            def join(self, timeout=None):
                super().join()

                return self.generated_ids

        streamer = (
            TextIteratorStreamer(  # TODO: Switch TextIteratorStreamer to AsyncTextIteratorStreamer
                # AsyncTextIteratorStreamer(
                self.processor,
                # decode_kwargs=dict(skip_special_tokens=True), # TODO: Why isn't this working?
                decode_kwargs={
                    "skip_special_tokens": True,
                },
                skip_prompt=True,
            )
            if stream
            else None
        )

        generation_kwargs = dict(
            # inputs, max_new_tokens=max_new_tokens, streamer=streamer, use_cache=True
            **inputs,
            max_new_tokens=max_new_tokens,
            streamer=streamer,
            use_cache=True,
        )

        thread = GeneratorThread(self.model, **generation_kwargs)
        thread.start()

        if stream:

            async def event_publisher():
                i = 0

                try:
                    for new_text in streamer:
                        # async for new_text in streamer:
                        choices: List[ChatCompletionChunkChoice] = [
                            ChatCompletionChunkChoice(
                                _request_id=None,
                                delta=ChoiceDelta(
                                    _request_id=None,
                                    content=new_text,
                                    function_call=None,
                                    refusal=None,
                                    role="assistant",
                                    tool_calls=None,
                                ),
                                finish_reason=None,
                                index=0,
                                logprobs=None,
                            )
                        ]

                        chat_completion_chunk = ChatCompletionChunk(
                            _request_id=None,
                            choices=choices,
                            created=int(time.time()),
                            id=str(i),
                            model=model_name,
                            object="chat.completion.chunk",
                            service_tier=None,
                            system_fingerprint=None,
                            usage=None,
                        )

                        yield chat_completion_chunk.model_dump_json()

                        i += 1

                except asyncio.CancelledError as e:
                    print("Disconnected from client (via refresh/close)")
                    # Do any other cleanup, if any
                    raise e
                except Exception as e:
                    print(f"Exception: {e}")
                    raise e

            return EventSourceResponse(event_publisher())

        generated_ids = thread.join()
        # input_length = inputs.shape[1]
        input_length = input_ids.shape[1]

        batch_decoded_outputs = self.processor.batch_decode(
            # generated_ids,
            generated_ids[:, input_length:],
            skip_special_tokens=True,
        )

        choices: List[ChatCompletionChoice] = []

        for i, response in enumerate(batch_decoded_outputs):
            # try:
            # response = json.loads(response)

            #         finish_reason: str = response.get("finish_reason")
            #         tool_calls_json = response.get("tool_calls")
            #         tool_calls: List[ToolCall] = []

            #         for tool_call_json in tool_calls_json:
            #             tool_call = ToolCall(
            #                 function=FunctionToolCallArguments(
            #                     arguments=tool_call_json.get("arguments"),
            #                     name=tool_call_json.get("name"),
            #                 ),
            #                 id=tool_call_json.get("id"),
            #                 type="function",
            #             )

            #             tool_calls.append(tool_call)

            #         message: ChatMessage = ChatMessage(
            #             role="assistant",
            #             tool_calls=tool_calls,
            #         )

            #     except json.JSONDecodeError:
            #         finish_reason: str = "stop"
            #         message: ChatMessage = ChatMessage(
            #             role="assistant",
            #             content=response,
            #         )

            message = ChatCompletionMessage(
                audio=None,
                content=response,
                refusal=None,
                role="assistant",
                tool_calls=None,
            )

            choices.append(
                ChatCompletionChoice(
                    index=i,
                    finish_reason="stop",
                    logprobs=None,
                    message=message,
                )
            )

        chat_completion = ChatCompletion(
            choices=choices,
            created=int(time.time()),
            id="1",
            model=model_name,
            object="chat.completion",
            service_tier=None,
            system_fingerprint=None,
            usage=None,
        )

        return chat_completion.model_dump(mode="json")


def build_app(cli_args: Dict[str, str]) -> serve.Application:
    """Builds the Serve app based on CLI arguments."""
    return VLLMDeployment.options().bind(
        cli_args.get("model_name"),
    )
